  #include "MinerOwnedStates.h"
#include "State.h"
#include "Miner.h"
#include "Locations.h"
#include "EntityNames.h"

#include <iostream>
using std::cout;
using std::ifstream;
using std::ofstream;
using std::endl;
 using std::string;
#include <cstring>


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//int estadoAnterior;
 
 #include "bhv_strict_check_shoot.h"

#include "GPR_ShootChance.h"
#include "body_kick_to_corner.h"
#include <rcsc/action/bhv_shoot.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

//define this to output to a file
#ifdef TEXTOUTPUT
#include <fstream>
extern std::ofstream os;
#define cout os
#endif


//----------------------------------------methods for SafePass

SafePass* SafePass::Instance()
{
    static SafePass instance;
    return &instance;
}

void SafePass::Enter(Miner* pMiner)
{
   // cout << "\n" << "          1: " << "Estou no Enter ";
}

void SafePass::Execute(Miner* pMiner )
{

    //cout << "\n" << "          1: " << "  EXECUTE.. ";

    rcsc::PlayerAgent * agent = pMiner->GetAgent();
    rcsc::dlog.addText( rcsc::Logger::TEAM,__FILE__": Bhv_BasicOffensiveKick" );
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
        ? static_cast< rcsc::PlayerObject * >( 0 )
        : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                           ? nearest_opp->pos()
                                           : rcsc::Vector2D( -1000.0, 0.0 ) );
    rcsc::Vector2D pass_point;

    //lancamento melhor passe
    if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL ) )
    {
       // cout << "\n" << "1 : " << "Safe PASS";
        if ( pass_point.x > wm.self().pos().x - 1.0 )
        {
            bool safety = true;
            const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
            for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
                    it != opps_end;
                    ++it )
            {
                if ( (*it)->pos().dist( pass_point ) < 4.0 )
                {
                    safety = false;
                }
            }

            if ( safety )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    __FILE__": (execute) do best pass" );
                agent->debugClient().addMessage( "OffKickPass(1)" );
                rcsc::Body_Pass().execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            }


        }
    }

    pMiner->ChangeState(NormalPass::Instance() );




}

void SafePass::Exit(Miner* pMiner)
{
    //cout << "\n" << "      1 : " << "EXIT"   ;
}



//--------------------------------------methods for AdvanceBall

AdvanceBall* AdvanceBall::Instance()
{
    static AdvanceBall instance;
    return &instance;
}

void AdvanceBall::Enter(Miner* pMiner)
{
    //cout << "\n" << " 2 : " << "ENTER                  ";
}

void AdvanceBall::Execute(Miner* pMiner)
{

    //cout << "\n" << " 2 : " << "Execute";

    rcsc::PlayerAgent * agent = pMiner->GetAgent();
    rcsc::dlog.addText( rcsc::Logger::TEAM, __FILE__": Bhv_BasicOffensiveKick");
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
        ? static_cast< rcsc::PlayerObject * >( 0 )
        : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                           ? nearest_opp->pos()
                                           : rcsc::Vector2D( -1000.0, 0.0 ) );
    const rcsc::Vector2D body_dir_drib_target
    = wm.self().pos()
      + rcsc::Vector2D::polar2vector(5.0, wm.self().body());
    rcsc::Vector2D pass_point;


    rcsc::Vector2D drib_target( 50.0, wm.self().pos().absY() );
    if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
    if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
    if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
    const rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();

    // opponent is behind of me
    if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
    {
        // check opponents
        // 15m, +-30 degree
        const rcsc::Sector2D sector( wm.self().pos(),
                                     0.5, 15.0,
                                     drib_angle - 30.0,
                                     drib_angle + 30.0 );
        // opponent check with goalie
        if ( ! wm.existOpponentIn( sector, 10, true ) )
        {
            const int max_dash_step
            = wm.self().playerType()
              .cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
            if ( wm.self().pos().x > 35.0 )
            {
                drib_target.y *= ( 10.0 / drib_target.absY() );
            }

            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": (execute) fast dribble to (%.1f, %.1f) max_step=%d",
                                drib_target.x, drib_target.y,
                                max_dash_step );
            agent->debugClient().addMessage( "OffKickDrib(2)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxDashPower(),
                                std::min( 5, max_dash_step )
                              ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": (execute) slow dribble to (%.1f, %.1f)",
                                drib_target.x, drib_target.y );
            agent->debugClient().addMessage( "OffKickDrib(3)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxDashPower(),
                                2
                              ).execute( agent );

        }

        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }


    

    //Advance Ball  // parametros mudados para avancar mais rapido

    // std::cout << agent->world().ball().pos().absX();


    //cout << "\n" << " 2 : " << "Advance BALL";

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__": (execute) opp far. dribble(%f, %f)",
                        drib_target.x, drib_target.x );
    agent->debugClient().addMessage( "OffKickDrib(10)" );
    rcsc::Body_Dribble( drib_target, 1.0, rcsc::ServerParam::i().maxDashPower() * 1.0, 1 ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
  


}
void AdvanceBall::Exit(Miner* pMiner)
{
   // cout << "\n" << "          2: " << "  EXIT ";

}

//--------------------------------------methods for NormalPass
NormalPass* NormalPass::Instance()
{
    static NormalPass instance;
    return &instance;
}

void NormalPass::Enter(Miner* pMiner)
{
    //cout << "\n" << "3 : " << "ENTER";
}

void NormalPass::Execute(Miner* pMiner)
{

   // cout << "\n" << "      3  : EXECUTE ";

    rcsc::PlayerAgent * agent = pMiner->GetAgent();
    rcsc::dlog.addText( rcsc::Logger::TEAM, __FILE__": Bhv_BasicOffensiveKick" );

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
        ? static_cast< rcsc::PlayerObject * >( 0 )
        : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                           ? nearest_opp->pos()
                                           : rcsc::Vector2D( -1000.0, 0.0 ) );

    rcsc::Vector2D pass_point;

    //passe normal

   // cout << "\n" << "3 : " << "Passe Normal";

    if ( rcsc::Body_Pass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (execute) do best pass" );
        agent->debugClient().addMessage( "OffKickPass(2)" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }


    else pMiner->ChangeState(Drible::Instance() );
}

void NormalPass::Exit(Miner* pMiner)
{
    //cout << "\n"  << "3 : " << "EXIT";
}



//------------------------------------------------methods for Drible

Drible* Drible::Instance()
{
    static Drible instance;
    return &instance;
}

void Drible::Enter(Miner* pMiner )
{
    //cout << "\n" << "  4 : " << "Enter  Drible  ";
}

void Drible::Execute(Miner* pMiner )
{



    //cout << "\n"  << "                4 : " << "EXECUTE   Drible     ";



    rcsc::PlayerAgent * agent = pMiner->GetAgent();

    rcsc::dlog.addText( rcsc::Logger::TEAM, __FILE__": Bhv_BasicOffensiveKick" );

    const rcsc::WorldModel & wm = agent->world();

    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();

    const rcsc::PlayerObject * nearest_opp

    = ( opps.empty()

        ? static_cast< rcsc::PlayerObject * >( 0 ) : opps.front() );

    const double nearest_opp_dist = ( nearest_opp

                                      ? nearest_opp->distFromSelf()

                                      : 1000.0 );

    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp

                                           ? nearest_opp->pos()

                                           : rcsc::Vector2D( -1000.0, 0.0 ) );

    const rcsc::Vector2D body_dir_drib_target

    = wm.self().pos()

      + rcsc::Vector2D::polar2vector(5.0, wm.self().body());

    rcsc::Vector2D pass_point;


    // dribble to my body dir

    /*   if ( nearest_opp_dist < 5.0

            && nearest_opp_dist > ( rcsc::ServerParam::i().tackleDist()

                                    + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 1.5 )

            && wm.self().body().abs() < 70.0 )

       {*/

    // const rcsc::Vector2D body_dir_drib_target

    //= wm.self().pos()

    // + rcsc::Vector2D::polar2vector(5.0, wm.self().body());



    int max_dir_count = 0;

    wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );



    if ( body_dir_drib_target.x < rcsc::ServerParam::i().pitchHalfLength() - 1.0

            && body_dir_drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0

            && max_dir_count < 3

       )

    {

        // check opponents

        // 10m, +-30 degree

        const rcsc::Sector2D sector( wm.self().pos(),

                                     0.5, 10.0,

                                     wm.self().body() - 30.0,

                                     wm.self().body() + 30.0 );

        // opponent check with goalie

        if ( ! wm.existOpponentIn( sector, 10, true ) )

        {

            rcsc::dlog.addText( rcsc::Logger::TEAM,

                                __FILE__": (execute) dribble to my body dir" );

            agent->debugClient().addMessage( "OffKickDrib(1)" );

            rcsc::Body_Dribble( body_dir_drib_target,

                                1.0,

                                rcsc::ServerParam::i().maxDashPower(),

                                2

                              ).execute( agent );

            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );



        }

    }



    rcsc::Vector2D drib_target( 50.0, wm.self().pos().absY() );

    if ( drib_target.y < 20.0 ) drib_target.y = 20.0;

    if ( drib_target.y > 29.0 ) drib_target.y = 27.0;

    if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;

    const rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();


    // opp is far from me

    if ( nearest_opp_dist > 2.6 )

    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,

                            __FILE__": (execute) opp far. dribble(%f, %f)",

                            drib_target.x, drib_target.y );

        agent->debugClient().addMessage( "OffKickDrib(5)" );

        rcsc::Body_Dribble( drib_target,

                            1.0,

                            rcsc::ServerParam::i().maxDashPower() * 0.2,

                            1

                          ).execute( agent );

        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }


    if ( nearest_opp_dist > 1.5 )

    {

        rcsc::dlog.addText( rcsc::Logger::TEAM,

                            __FILE__": hold" );

        agent->debugClient().addMessage( "OffKickHold" );

        rcsc::Body_HoldBall().execute( agent );

        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

    }


    else if ( wm.self().pos().x > wm.offsideLineX() - 10.0 )

    {

        rcsc::dlog.addText( rcsc::Logger::TEAM,

                            __FILE__": kick near side" );

        agent->debugClient().addMessage( "OffKickToCorner" );

        Body_KickToCorner( (wm.self().pos().y < 0.0) ).execute( agent );

        agent->setNeckAction( new rcsc::Neck_ScanField() );

    }

  /*  else

    {

        rcsc::dlog.addText( rcsc::Logger::TEAM,

                            __FILE__": clear" );

        agent->debugClient().addMessage( "OffKickAdvance" );

        rcsc::Body_AdvanceBall().execute( agent );

        agent->setNeckAction( new rcsc::Neck_ScanField() );

    }*/


}

void Drible::Exit(Miner* pMiner )
{
    //cout << "\n"  << " 4  : " << "  :  EXITE Drible        ";
}


//----------------------------------------methods for StateBegin

StateBegin* StateBegin::Instance()
{
    static StateBegin instance;
    return &instance;
}

void StateBegin::Enter(Miner* pMiner)
{
   // cout << "\n" << "          0: " << "Estou no Enter ";
}

void StateBegin::Execute(Miner* pMiner )
{
    rcsc::PlayerAgent * agent = pMiner->GetAgent();
    rcsc::dlog.addText( rcsc::Logger::TEAM,__FILE__": Bhv_BasicOffensiveKick" );
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
        ? static_cast< rcsc::PlayerObject * >( 0 )
        : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                           ? nearest_opp->pos()
                                           : rcsc::Vector2D( -1000.0, 0.0 ) );
    rcsc::Vector2D pass_point;

    int matrizQ[7][7];
    int matrizR[7][7];

    static int estadoAnterior;
    static int acaoAnterior;
    int estado = 0;
    int acao = 0;
    float alpha = 0.5;
    float gama = 0.8;
    
    

    float q = 0.0;
    float E = 0.2;
    float teste = 0;
    srand(time(NULL));
    //float vet[7];
    
    // std::string foo = "hi";
 
  // Now we can just say "string".
  char matrizq[13] = "matrizQ1.txt";
  char matrizr[13] = "matrizR1.txt";
  
 
   

    
    switch ( pMiner->getId() ) {


    case 7:
      matrizq[7] = '3';
      matrizr[7] = '3';
        break;
    case 8:
      matrizq[7] = '2';
      matrizr[7] = '2';
        break;
     case 9:
      matrizq[7] = '1';
      matrizr[7] = '1';
        break;
       default:
          break;
    }
    
  

    //carrega matriQ
    ifstream iMatrizQ ( matrizq);
    while ( iMatrizQ  && !iMatrizQ.eof()) {
        for (int i = 0; i < 7; i++) {
            for (int j = 0; j < 7; j++) {
                iMatrizQ>>matrizQ[i][j];
            }
        }
    }
    iMatrizQ.close();

     
    //carrega matriR
    ifstream iMatrizR ( matrizr );
    while ( iMatrizR  && !iMatrizR.eof()) {
        for (int i = 0; i < 7; i++) {
            for (int j = 0; j < 7; j++) {
                iMatrizR>>matrizR[i][j];
            }
        }
    }
    iMatrizR.close();
    
    
 
     // opponent is behind of me
    if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 ) //1.0
    {
        estado = 1;
        //cout << "\n"<<estado<< " oponente atraz";
    }
    //passe normal
    else if ( nearest_opp_dist > 6.0 ) //era 7
    {
        estado = 2; //cout << "\n"<<*ptEstado<< " adversario longe";
    }
    // opp is far from me //Advance Ball
    else if ( nearest_opp_dist > 5.0 )
    {
        estado = 3;
        //cout << "\n"<<estado<<" adversario medio";
    }
    // opp is far from me
    else if ( nearest_opp_dist > 3.0 )
    {
        estado = 4;
        //cout << "\n"<<estado<< " adversario perto";

    }
    else if ( nearest_opp_dist <= 3.0 )
    {
        estado = 5;
        //cout << "\n"<<estado<<" adversario muito perto";
    }

    if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL ) )
    {
        if ( pass_point.x > wm.self().pos().x - 1.0 )
        {
            bool safety = true;
            const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
            for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
                    it != opps_end;
                    ++it )
            {
                if ( (*it)->pos().dist( pass_point ) < 4.0 )
                {
                    safety = false;
                }
            }

            if ( safety )
            {

                estado = 0;
                //cout << "\n" <<estado<<" lancamento livre";
            }
        }
    }
    
    
  //  const WorldModel & wm = agent->world();

    if (  agent->world().gameMode().type() != rcsc::GameMode::IndFreeKick_
         &&  agent->world().time().stopped() == 0
         &&  agent->world().self().isKickable()
         && Bhv_StrictCheckShoot().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                      __FILE__": shooted" );

        // reset intention
        agent->setIntention( static_cast< rcsc::SoccerIntention * >( 0 ) );
        //return true;
	
	estado = 6;
    }

    //return false;
    
    // check shoot chance
   /* if ( agent->world().gameMode().type() != rcsc::GameMode::IndFreeKick_
            && agent->world().self().isKickable()
            && rcsc::Bhv_Shoot2008().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": shooted" );

        // reset intention
        agent->setIntention( static_cast< rcsc::SoccerIntention * >( 0 ) );

        estado = 6;


    }*/



    q = (rand()%10) ;
    q = q /10;
    //cout<<"\n"<<"q : "<<q;


    if (q <= E) { //escolha uma acao aleatoria

        /*int count = 0;
        for ( int i = 0; i < 6; i++ ) {
         if(matrizR[estado][i] != -1) {
             vet[count] = matrizR[estado][i];
             count++;
         }
         acao = rand()%count;
         }
         acao = vet[acao];*/
        //solucao anterior
        do { //encontrar uma solucao melhor depois
            acao = rand()%7;
            //cout<<"\n"<<"acao : "<<acao;
            //cout<<"\n"<<matrizR[estado][acao];
        } while (matrizR[estado][acao] == -1);
    }
    else { // encontre acao maximo
        int max1 = -1;

        for ( int i = 0; i < 7; i++ ) {

            if (matrizR[estado][i] != -1) {

                if (matrizQ[estado][i] > max1) {
                    max1 = matrizQ[estado][i];
                    acao = i;
                }
            }
        }
        //cout<<"\n"<<"max1 : "<<max1;
    }

  /*  cout<<"\n"<<"estado : "<<estado;
    cout<<"\n"<<"acao : "<<acao;
    cout<<"\n";
    cout<<"\n"<<"estadoAnterior : "<<estadoAnterior;
    cout<<"\n"<<"acaoAnterior : "<<acaoAnterior;
    cout<<"\n";
    cout<<"\n";*/




    //se houver mudanca de estado com relacao ao estado anterior atualiza a matrizQ
//com o valor do reforco utilizado


    if (estado != estadoAnterior) {

        int max2 = 0;



        for ( int i = 0; i < 7; i++ ) {

            if (matrizQ[estado][i] >= max2) {
                max2 = matrizQ[estado][i];
                //cout<<"\n";
                //cout<<matrizQ[estado][i];

            }
        }

       // cout<<"\n"<<"max2 : "<<max2;
       // cout<<"\n";
       // cout<<"\n";

       // cout<<matrizQ[estadoAnterior][acaoAnterior]<<" + "<<alpha<<" * ("<<matrizR[estadoAnterior][acaoAnterior]<<" + "<<gama<<" * "<<max2<<" - "<<matrizQ[estadoAnterior][acaoAnterior]<<" ) ";
        //max = máximo do estado atual
        teste =  matrizQ[estadoAnterior][acaoAnterior] + alpha * (matrizR[estadoAnterior][acaoAnterior] + gama* max2 -  matrizQ[estadoAnterior][acaoAnterior] );

        matrizQ[estadoAnterior][acaoAnterior] = teste;

       // cout<<"\n";
       // cout<<"\n"<<"teste : "<<teste;
       // cout<<"\n";
       // cout<<"\n";

        char espaco = ' ';
        ofstream iMatrizQS(matrizq );

        for (int i = 0; i < 7; i++) {
            for (int j = 0; j < 7; j++) {
                iMatrizQS<< matrizQ[i][j]<<espaco;
            }
            iMatrizQS<<endl;
        }
        iMatrizQS.close();

      //  for (int i = 0; i < 7; i++) {
      //      for (int j = 0; j < 7; j++) {
      //          cout <<" "<<matrizQ[i][j];
      //      }
           // cout <<"\n";
      //  }
//
      // cout <<"\n";



    }


    switch ( acao ) {

    case 0:
        //pMiner->ChangeState(SafePass::Instance() );
        pMiner->ChangeState(NormalPass::Instance() );
        break;
    case 1:
        pMiner->ChangeState(AdvanceBall::Instance() );
        break;
    case 2:
        pMiner->ChangeState(AdvanceBall::Instance() );
        break;
    case 3:
        pMiner->ChangeState(NormalPass::Instance() );
        break;
    case 4:
        pMiner->ChangeState(Drible::Instance() );
        break;
    case 5:
        pMiner->ChangeState(Drible::Instance() );
        break;

    default:

        // pMiner->ChangeState(AdvanceBall::Instance() );

        break;
    }


    estadoAnterior = estado;
    acaoAnterior = acao;

}

void StateBegin::Exit(Miner* pMiner)
{
    //cout << "\n" << "      1 : " << "EXIT"   ;
}
